"""
This file contains the Assertion class, which is used to represent an assertion in DataHub.

The Assertion class is currently not implemented, this is a placeholder for future implementation.
"""

from typing import Union

from datahub.metadata.urns import AssertionUrn


class Assertion:
    def __init__(self, urn: Union[str, AssertionUrn]):
        print(f"The Assertion class is currently not implemented. Urn provided: {urn}")
        self.urn = AssertionUrn(urn)



import logging
import time
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import BaseModel
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from acryl_datahub_cloud.datahub_metadata_sharing.query import (
    GRAPHQL_SCROLL_SHARED_ENTITIES,
    GRAPHQL_SHARE_ENTITY,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph

logger = logging.getLogger(__name__)


class GraphQLError(Exception):
    """Custom exception for GraphQL-specific errors"""

    pass


class DataHubMetadataSharingSourceConfig(BaseModel):
    batch_size: int = 100
    batch_delay_ms: int = 100
    max_retries: int = 3
    initial_retry_delay_ms: int = 1000


@dataclass
class DataHubMetadataSharingSourceReport(SourceReport):
    entities_shared: int = 0
    entities_failed: int = 0
    implicit_entities_skipped: int = 0
    batches_processed: int = 0


@platform_name(id="datahub", platform_name="DataHub")
@config_class(DataHubMetadataSharingSourceConfig)
@support_status(SupportStatus.INCUBATING)
class DataHubMetadataSharingSource(Source):
    """MetadataSharing Source that reshares entities across DataHub instances"""

    def __init__(
        self, config: DataHubMetadataSharingSourceConfig, ctx: PipelineContext
    ):
        super().__init__(ctx)
        self.config: DataHubMetadataSharingSourceConfig = config
        self.report = DataHubMetadataSharingSourceReport()
        self.graph: Optional[DataHubGraph] = None

    @retry(
        retry=retry_if_exception_type((GraphQLError, ConnectionError)),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        reraise=True,
    )
    def execute_graphql_with_retry(
        self, query: str, variables: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Execute GraphQL query with retry logic"""
        if self.graph is None:
            raise ValueError("Graph client not initialized")
        response = self.graph.execute_graphql(query, variables=variables)
        error = response.get("error")
        if error:
            raise GraphQLError(f"GraphQL error: {error}")
        return response

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        self.graph = self.ctx.require_graph("Loading default graph coordinates.")

        self.reshare_entities()

        # This source doesn't produce any work units
        return []

    def reshare_entities(self) -> None:
        scroll_id: Optional[str] = None
        current_batch_number: int = 1

        try:
            while True:
                next_scroll_id, results = self.scroll_shared_entities(
                    scroll_id, self.config.batch_size
                )

                for result in results:
                    self._process_single_entity(result)

                self.report.batches_processed = current_batch_number
                logger.info(
                    f"Completed sharing batch of entities {current_batch_number} of size {self.config.batch_size}!"
                )
                current_batch_number += 1

                if next_scroll_id is None:
                    break
                else:
                    scroll_id = next_scroll_id

                time.sleep(self.config.batch_delay_ms / 1000.0)

        except Exception as e:
            self.report.report_failure(
                title="Failed to process batches",
                message="Error occurred while processing one or more batches!",
                context=f"message = {str(e)}",
                exc=e,
            )
            return

        logger.info(
            f"Successfully shared {self.report.entities_shared} entities, failed to share {self.report.entities_failed} entities.",
        )

    # Rest of the methods remain the same...

    def _process_single_entity(self, result: Dict[str, Any]) -> None:
        """Process a single entity result"""
        entity_urn = result.get("entity", {}).get("urn", None)
        share_results = (
            result.get("entity", {}).get("share", {}).get("lastShareResults", [])
        )

        if entity_urn is None:
            self.report.report_warning(
                message="Failed to resolve entity urn for shared asset! Skipping...",
                context=f"Response: {str(result)}",
            )
            return

        for share_result in share_results:
            try:
                destination_data = share_result.get("destination", {})
                destination_urn = destination_data.get("urn", "")
                previous_status = share_result.get("status")
                share_config = share_result.get("shareConfig", {})

                # Important: If there is implicit entity, we should skip this urn.
                # This means the entity was not EXPLICITLY shared, so we do not want to explicitly share here.
                implicit_shared_entity = share_result.get("implicitShareEntity")
                is_implicitly_shared = (
                    implicit_shared_entity is not None
                    and "urn" in implicit_shared_entity
                )

                if is_implicitly_shared:
                    self.report.implicit_entities_skipped += 1
                    continue

                if previous_status != "SUCCESS":
                    logger.info(
                        "Attempting to share a previously unsuccessful shared entity! "
                        + f"entity urn: {entity_urn}, destination urn: {destination_urn}"
                    )

                lineage_direction = self._determine_lineage_direction(share_config)

                shared = self.share_entity(
                    entity_urn=entity_urn,
                    destination_urn=destination_urn,
                    lineage_direction=lineage_direction,
                )

                if shared:
                    self.report.entities_shared += 1
                else:
                    self.report.entities_failed += 1

            except Exception as e:
                self.report.report_warning(
                    message="Failed to share single entity!",
                    context=f"entity urn: {entity_urn}",
                )
                logger.exception(f"Error processing entity {entity_urn}", e)
                self.report.entities_failed += 1

    def _determine_lineage_direction(
        self, share_config: Optional[Dict[str, Any]]
    ) -> Optional[str]:
        if share_config is None:
            return None

        """Determine lineage direction based on share config"""
        include_upstreams = share_config.get("enableUpstreamLineage", False)
        include_downstreams = share_config.get(
            "enableDownstreamLineage", False
        )  # Fixed typo

        if include_upstreams and include_downstreams:
            return "BOTH"
        if include_upstreams:
            return "UPSTREAM"
        if include_downstreams:
            return "DOWNSTREAM"
        return None

    def scroll_shared_entities(
        self, scroll_id: Optional[str], count: int
    ) -> Tuple[Optional[str], List[Dict[str, Any]]]:
        """Scroll through shared entities with retry logic"""
        response = self.execute_graphql_with_retry(
            GRAPHQL_SCROLL_SHARED_ENTITIES,
            variables={
                "scrollId": scroll_id,
                "count": count,
            },
        )

        result = response.get("scrollAcrossEntities", {})
        return result.get("nextScrollId"), result.get("searchResults", [])

    def share_entity(
        self, entity_urn: str, destination_urn: str, lineage_direction: Optional[str]
    ) -> bool:
        """Share entity with retry logic"""
        try:
            response = self.execute_graphql_with_retry(
                GRAPHQL_SHARE_ENTITY,
                variables={
                    "entityUrn": entity_urn,
                    "destinationUrn": destination_urn,
                    "lineageDirection": lineage_direction,
                },
            )

            result = response.get("shareEntity", {})
            if not result.get("succeeded", False):
                self.report.report_failure(
                    title="Failed to Share Entity",
                    message="Response returned that success failed for entity and destination!",
                    context=f"entity urn: {entity_urn}, destination urn: {destination_urn}",
                )
                return False

            return True

        except Exception as e:
            self.report.report_failure(
                title="Failed to Share Entity",
                message="Exception occurred while sharing entity",
                context=f"entity urn: {entity_urn}, destination urn: {destination_urn}",
                exc=e,
            )
            return False

    def get_report(self) -> SourceReport:
        return self.report



import logging
import os
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

import boto3
from opensearchpy import OpenSearch
from pydantic import validator

from acryl_datahub_cloud.datahub_reporting.datahub_dataset import (
    DataHubBasedS3Dataset,
    DatasetMetadata,
    DatasetRegistrationSpec,
    FileStoreBackedDatasetConfig,
)
from acryl_datahub_cloud.elasticsearch.config import ElasticSearchClientConfig
from acryl_datahub_cloud.elasticsearch.graph_service import ElasticGraphRow
from datahub.configuration.common import ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.graph.client import DatahubClientConfig, DataHubGraph
from datahub.metadata.schema_classes import DatasetPropertiesClass

logger = logging.getLogger(__name__)


class DataHubReportingExtractGraphSourceConfig(ConfigModel):
    enabled: bool = True
    server: Optional[DatahubClientConfig] = None
    search_index: ElasticSearchClientConfig = ElasticSearchClientConfig()
    extract_graph_store: FileStoreBackedDatasetConfig
    relationships_include: Optional[List[str]] = None
    relationships_exclude: Optional[List[str]] = None
    entity_types_include: Optional[List[str]] = None
    entity_types_exclude: Optional[List[str]] = None
    query_timeout: int = 30
    extract_batch_size: int = 2000

    @validator("extract_graph_store", pre=True, always=True)
    def set_default_extract_soft_delete_flag(cls, v, values):
        if v is not None:
            if "dataset_registration_spec" not in v:
                v["dataset_registration_spec"] = DatasetRegistrationSpec(
                    soft_deleted=False
                )
            elif "soft_deleted" not in v["dataset_registration_spec"]:
                v["dataset_registration_spec"]["soft_deleted"] = False
        return v


@dataclass
class DataHubReportingExtractGraphSourceReport(SourceReport):
    edges_scanned: int = 0
    edges_skipped: int = 0
    file_store_connection_status: str = "Not connected"
    file_store_uri: str = ""

    def increment_edges_scanned(self):
        self.edges_scanned += 1

    def increment_edges_skipped(self):
        self.edges_skipped += 1

    def as_string(self) -> str:
        return super().as_string()


@platform_name(id="datahub", platform_name="DataHub")
@config_class(DataHubReportingExtractGraphSourceConfig)
@support_status(SupportStatus.INCUBATING)
class DataHubReportingExtractGraphSource(Source):
    platform = "datahub"

    def __init__(
        self, config: DataHubReportingExtractGraphSourceConfig, ctx: PipelineContext
    ):
        super().__init__(ctx)
        self.config: DataHubReportingExtractGraphSourceConfig = config
        self.report = DataHubReportingExtractGraphSourceReport()
        self.opened_files: List[str] = []
        self.s3_client = boto3.client("s3")
        self.datahub_based_s3_dataset = DataHubBasedS3Dataset(
            self.config.extract_graph_store,
            dataset_metadata=DatasetMetadata(
                displayName="Graph Extract",
                description="This is an automated Graph Extract from DataHub's backend",
            ),
        )

    def process_batch(self, results):
        for doc in results:
            row = ElasticGraphRow.from_elastic_doc(doc["_source"])
            self.report.increment_edges_scanned()
            self.datahub_based_s3_dataset.append(row)

    def should_skip_graph_extract(
        self, dataset_properties: Optional[DatasetPropertiesClass]
    ) -> bool:
        """Skip graph extraction if the dataset has already been pushed to DataHub today"""

        skip_extract = False
        # If present check if the updated timestamp is for the current day
        if dataset_properties and dataset_properties.lastModified:
            # datetime function requires timestamp in seconds
            ts = datetime.fromtimestamp(dataset_properties.lastModified.time / 1000)
            skip_extract = ts.day is datetime.today().day

        if skip_extract:
            logger.info(
                f"Skipping graph extract as dataset has been updated today {ts}"
            )
        return skip_extract

    def get_workunits(self):
        if not self.config.enabled:
            logger.info("Source is disabled, stopping")
            return

        self.graph = (
            self.ctx.require_graph("Loading default graph coordinates.")
            if self.config.server is None
            else DataHubGraph(config=self.config.server)
        )

        dataset_properties: Optional[DatasetPropertiesClass] = self.graph.get_aspect(
            self.datahub_based_s3_dataset.get_dataset_urn(), DatasetPropertiesClass
        )

        if (
            self.should_skip_graph_extract(dataset_properties)
            and self.datahub_based_s3_dataset.config.generate_presigned_url
        ):
            mcps = self.datahub_based_s3_dataset.update_presigned_url(
                dataset_properties=dataset_properties
            )
            for mcp in mcps:
                yield mcp.as_workunit()
        else:
            endpoint = self.config.search_index.endpoint
            index_prefix = self.config.search_index.index_prefix
            user = self.config.search_index.username
            password = self.config.search_index.password
            batch_size = self.config.extract_batch_size
            server = OpenSearch(
                [endpoint],
                http_auth=(user, password),
                use_ssl=self.config.search_index and self.config.search_index.use_ssl,
            )

            query = {
                "query": {"match_all": {}},
                "sort": [
                    {"source.urn": {"order": "desc"}},
                    {"destination.urn": {"order": "desc"}},
                    {"relationshipType": {"order": "desc"}},
                    {"lifecycleOwner": {"order": "desc"}},
                ],
            }

            index = f"{index_prefix}graph_service_v1"
            response = server.create_pit(index, keep_alive="10m")

            # TODO: Save PIT, we can resume processing based on <pit, search_after> tuple
            pit = response.get("pit_id")
            query.update({"pit": {"id": pit, "keep_alive": "10m"}})

            # TODO: Using slicing we can parallelize the ES calls below:
            # https://opensearch.org/docs/latest/search-plugins/searching-data/point-in-time/#search-slicing
            while True:
                results = server.search(
                    body=query,
                    size=batch_size,
                    params={"timeout": self.config.query_timeout},
                )
                self.process_batch(results["hits"]["hits"])
                if len(results["hits"]["hits"]) < batch_size:
                    break
                query.update({"search_after": results["hits"]["hits"][-1]["sort"]})

            response = server.delete_pit(body={"pit_id": pit})

            self.datahub_based_s3_dataset.commit()

            for mcp in self.datahub_based_s3_dataset.commit():
                self.report.file_store_uri = (
                    self.datahub_based_s3_dataset.get_file_uri()
                )
                logger.info(f"Reporting file created at {self.report.file_store_uri}")
                logger.info(
                    f"Reporting dataset registered at {self.datahub_based_s3_dataset.get_dataset_urn()}"
                )
                yield mcp.as_workunit()

    def get_report(self) -> SourceReport:
        return self.report

    def close(self) -> None:
        for file in self.opened_files:
            os.remove(file)
        return super().close()



import pathlib

GRAPHQL_SCROLL_SHARED_ENTITIES = (
    pathlib.Path(__file__).parent / "scroll_shared_entities.gql"
).read_text()

GRAPHQL_SHARE_ENTITY = (pathlib.Path(__file__).parent / "share_entity.gql").read_text()
